/*
Code is released under Simplified BSD License.
-------------------------------------------------------------------------------
Copyright 2012 Nghia Ho. All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are
permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of
      conditions and the following disclaimer.

   2. Redistributions in binary form must reproduce the above copyright notice, this list
      of conditions and the following disclaimer in the documentation and/or other materials
      provided with the distribution.

THIS SOFTWARE IS PROVIDED BY NGHIA HO ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NGHIA HO OR
CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those of the
authors and should not be interpreted as representing official policies, either expressed
or implied, of Nghia Ho.
*/
#include <GL/freeglut.h>
#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <sstream>
#include <cassert>

#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgproc/imgproc.hpp>

#include "Camera.h"

using namespace std;

const int DOWNSAMPLE_WIDTH = 512;

struct Point3D
{
    float x, y, z;
};

struct Face
{
    float x[3], y[3], z[3];
    float u[3], v[3]; // UV co-ordinates, normalised to [0..1]
    int image_num; // best image assignment
};

struct RGB
{
    unsigned char r, g, b;
};

void Display();
void Reshape (int w, int h);
void Keyboard(unsigned char key, int x, int y);
void KeyboardUp(unsigned char key, int x, int y);
void Special(int key, int x, int y);
void SpecialUp(int key, int x, int y);
void MouseMotion(int x, int y);
void Mouse(int button, int state, int x, int y);
void Timer(int value);
void Idle();
void Enable();

void LoadFaces(const std::string &filename, std::vector <Face> &faces, std::vector <cv::Mat> &images);
void LoadPoints(const std::string &filename, std::vector <Point3D> &points);
void RenderBitmapString(float x, float y, float z, void *font, char *string);
void Grid();
void DrawHelp();

string g_dir;
Camera g_camera;
bool g_key[256];
bool g_special[256];
bool g_shift_down = false;
bool g_fps_mode = false;
int g_viewport_width = 0;
int g_viewport_height = 0;
bool g_mouse_left_down = false;
bool g_mouse_right_down = false;
bool g_invert_mouse = true;
bool g_show_sphere = true;
bool g_show_help = true;

enum DisplayMode {POINTCLOUD, WIREFRAME, NORMAL};
DisplayMode g_display_mode = NORMAL;
GLuint g_display_list[3] = {0, 0, 0};

// Movement settings
const float g_translation_speed = 0.05;
const float g_rotation_speed = M_PI/180*0.2;

vector <Face> g_faces;
vector <cv::Mat> g_images;
vector <GLuint> g_tex;
vector <RGB> g_colours;
vector <Point3D> g_points;

int main (int argc, char **argv) {
    if(argc < 2) {
        cout << "./prog [pmvs_dir]" << endl;
        return 0;
    }

    g_dir = argv[1];

    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize(800, 600);
    glutCreateWindow("ViewMesh by Nghia Ho - Hit SPACEBAR to enter navigation mode");

    Enable();

    LoadFaces(g_dir + "/models/output.mesh", g_faces, g_images);
    LoadPoints(g_dir + "/models/output.point", g_points);

    glutIgnoreKeyRepeat(1);

    glutDisplayFunc(Display);
    glutIdleFunc(Display);
    glutReshapeFunc(Reshape);
    glutMouseFunc(Mouse);
    glutMotionFunc(MouseMotion);
    glutPassiveMotionFunc(MouseMotion);
    glutKeyboardFunc(Keyboard);
    glutKeyboardUpFunc(KeyboardUp);
    glutSpecialFunc(Special);
    glutSpecialUpFunc(SpecialUp);
    glutIdleFunc(Idle);
    glutTimerFunc(1, Timer, 0);

    glutMainLoop();

    return 0;
}

void LoadPoints(const std::string &filename, std::vector <Point3D> &points)
{
    ifstream input(filename.c_str());

    if(!input) {
        cerr << "Error opening " << filename << endl;
        exit(1);
    }

    size_t num;
    char line[1024];

    input.getline(line, sizeof(line));
    stringstream str(line);
    str >> num;

    points.resize(num);

    input.read((char*)&points[0], sizeof(Point3D)*num);
}

void LoadFaces(const std::string &filename, std::vector <Face> &faces, std::vector <cv::Mat> &images)
{
    ifstream input(filename.c_str());

    if(!input) {
        cerr << "Error opening " << filename << endl;
        exit(1);
    }

    size_t num_faces, num_images;

    input >> num_faces >> num_images;

    g_colours.resize(num_images);

    for(size_t i=0; i < num_images; i++) {
        g_colours[i].r = 55 + 200*(rand()/(1.0 + RAND_MAX));
        g_colours[i].g = 55 + 200*(rand()/(1.0 + RAND_MAX));
        g_colours[i].b = 55 + 200*(rand()/(1.0 + RAND_MAX));
    }

    g_tex.resize(num_images);
    glGenTextures(num_images, &g_tex[0]);
    assert(glGetError() == GL_NO_ERROR);

    for(size_t i=0; i < num_images; i++) {
        char file[1024];

        // Check if downsample exist
        sprintf(file, "%s/visualize/%08ds.jpg", g_dir.c_str(), (int)i);

        cv::Mat img = cv::imread(file);

        if(!img.data) {
            // Check if original image exist
            sprintf(file, "%s/visualize/%08d.jpg", g_dir.c_str(), (int)i);
            img = cv::imread(file);

            if(!img.data) {
                cerr << "Can't find " << file << endl;
                exit(1);
            }

            if(img.cols > DOWNSAMPLE_WIDTH) {
                int new_width = DOWNSAMPLE_WIDTH;
                int new_height = img.rows * DOWNSAMPLE_WIDTH / img.cols;

                cv::resize(img, img, cv::Size(new_width, new_height));

                sprintf(file, "%s/visualize/%08ds.jpg", g_dir.c_str(), (int)i);
                cv::imwrite(file, img);

                cout << "Creating down sample image" << endl;
            }
        }

        cout << "Loading " << file << endl;

        cv::flip(img, img, 0);
        glBindTexture(GL_TEXTURE_2D, g_tex[i]);
        assert(glGetError() == GL_NO_ERROR);

        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img.cols, img.rows, 0, GL_RGB, GL_UNSIGNED_BYTE, img.data);
        assert(glGetError() == GL_NO_ERROR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    }

    faces.resize(num_faces);
    for(size_t i=0; i < num_faces; i++) {
        for(int j=0; j < 3; j++) {
            input >> faces[i].x[j];
            input >> faces[i].y[j];
            input >> faces[i].z[j];
        }

        input >> faces[i].image_num;

        for(int j=0; j < 3; j++) {
            input >> faces[i].u[j];
            input >> faces[i].v[j];
        }
    }
}

void Grid()
{
    glPushMatrix();
    glColor3f(1,1,1);

    for(int i=-50; i < 50; i++) {
        glBegin(GL_LINES);
        glVertex3f(i, 0, -50);
        glVertex3f(i, 0, 50);
        glEnd();
    }

    for(int i=-50; i < 50; i++) {
        glBegin(GL_LINES);
        glVertex3f(-50, 0, i);
        glVertex3f(50, 0, i);
        glEnd();
    }

    glPopMatrix();
}

void Enable() {
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST); //enable the depth testing

    glEnableClientState(GL_VERTEX_ARRAY);

   // glEnableClientState(GL_COLOR_ARRAY);
    //glEnable (GL_LIGHTING); //enable the lighting
    //glEnable (GL_LIGHT0); //enable LIGHT0, our Diffuse Light
    //glShadeModel (GL_SMOOTH); //set the shader to smooth shader
}

void Display() {
    static int frame_count = 0;
    static int last_time = glutGet(GLUT_ELAPSED_TIME);

    glClearColor (0.0,0.0,0.0,1.0); //clear the screen to black
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT); //clear the color buffer and the depth buffer
    glLoadIdentity();

	g_camera.Refresh();

    glColor3f(1,1,1);

    // Create a display list if it doesn't exist already
    if(g_display_list[g_display_mode] == 0) {
        g_display_list[g_display_mode] = glGenLists(1);
        glNewList(g_display_list[g_display_mode], GL_COMPILE);

        if(g_display_mode == NORMAL) {
            glEnable(GL_TEXTURE_2D);
            for(size_t i=0; i < g_faces.size(); i++) {
                glBindTexture(GL_TEXTURE_2D, g_tex[g_faces[i].image_num]);
                assert(glGetError() == GL_NO_ERROR);

                glBegin(GL_TRIANGLES);
                for(int j=0; j < 3; j++) {
                    float u = g_faces[i].u[j];
                    float v = g_faces[i].v[j];
                    float x = g_faces[i].x[j];
                    float y = g_faces[i].y[j];
                    float z = g_faces[i].z[j];

                    glTexCoord2f(u,v); glVertex3f(x,y,z);
                }
                glEnd();

                assert(glGetError() == GL_NO_ERROR);
            }
            glDisable(GL_TEXTURE_2D);
        }
        else if(g_display_mode == POINTCLOUD) {
            glColor3f(0,0.8,0);
            glVertexPointer(3, GL_FLOAT, 0, &g_points[0]);
            glDrawArrays(GL_POINTS, 0, g_points.size());
        }
        else if(g_display_mode == WIREFRAME) {

            for(size_t i=0; i < g_faces.size(); i++) {
                int idx = g_faces[i].image_num;
                glColor3ub(g_colours[idx].r, g_colours[idx].g, g_colours[idx].b);

                glBegin(GL_LINE_LOOP);
                for(int j=0; j < 3; j++) {
                    float x = g_faces[i].x[j];
                    float y = g_faces[i].y[j];
                    float z = g_faces[i].z[j];

                    glVertex3f(x,y,z);
                }
                glEnd();
            }
        }
        glEndList();
    }
    else {
        glCallList(g_display_list[g_display_mode]);
    }

    //glutWireTeapot(0.2);
    //Grid();

    if(g_show_sphere) {
        glColor3f(1.0f, 1.0f, 1.0f);
        glutWireSphere(100.0, 20, 20);
    }

    if(g_show_help) {
        DrawHelp();
    }

    glutSwapBuffers(); //swap the buffers

    frame_count++;

    if(frame_count == 100) {
        int t = glutGet(GLUT_ELAPSED_TIME);
        int delta = t - last_time;
        float fps = 100*1000.0f/delta;

        //cout << fps << endl;

        last_time = t;
        frame_count = 0;
    }
}

void Reshape (int w, int h) {
    g_viewport_width = w;
    g_viewport_height = h;

    glViewport (0, 0, (GLsizei)w, (GLsizei)h); //set the viewport to the current window specifications
    glMatrixMode (GL_PROJECTION); //set the matrix to projection

    glLoadIdentity ();
    gluPerspective (60, (GLfloat)w / (GLfloat)h, 0.1 , 200.0); //set the perspective (angle of sight, width, height, ,depth)
    glMatrixMode (GL_MODELVIEW); //set the matrix back to model
}

void Keyboard(unsigned char key, int x, int y)
{
    switch(key) {
        case 27:
        exit(0);
        break;

        case ' ': {
            g_fps_mode = !g_fps_mode;

            if(g_fps_mode) {
                glutSetCursor(GLUT_CURSOR_NONE);
                glutWarpPointer(g_viewport_width/2, g_viewport_height/2);
            }
            else {
                glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
            }
        }
        break;

        case '1':
        g_display_mode = POINTCLOUD;
        break;

        case '2':
        g_display_mode = WIREFRAME;
        break;

        case '3':
        g_display_mode = NORMAL;
        break;

        case 'i':
        g_invert_mouse = !g_invert_mouse;
        break;

        case 't':
        g_show_sphere = !g_show_sphere;
        break;

        case 'h':
        g_show_help = !g_show_help;
        break;
    }

    if(glutGetModifiers() == GLUT_ACTIVE_SHIFT) {
        g_shift_down = true;
    }
    else {
        g_shift_down = false;
    }

    g_key[key] = true;
}

void KeyboardUp(unsigned char key, int x, int y)
{
    g_key[key] = false;
}

void Special(int key, int x, int y)
{
    if(glutGetModifiers() == GLUT_ACTIVE_SHIFT) {
        g_shift_down = true;
    }
    else {
        g_shift_down = false;
    }

    g_special[key] = true;
}

void SpecialUp(int key, int x, int y)
{
    g_special[key] = false;
}

void Timer(int value)
{
    if(g_key['w'] || g_key['W']) {
        g_camera.Move(g_translation_speed);
    }
    else if(g_key['s'] || g_key['S']) {
        g_camera.Move(-g_translation_speed);
    }
    else if(g_key['a'] || g_key['A']) {
        g_camera.Strafe(g_translation_speed);
    }
    else if(g_key['d'] || g_key['D']) {
        g_camera.Strafe(-g_translation_speed);
    }
    else if(g_mouse_left_down) {
        g_camera.Fly(-g_translation_speed);
    }
    else if(g_mouse_right_down) {
        g_camera.Fly(g_translation_speed);
    }

    glutTimerFunc(1, Timer, 0);
}

void Idle()
{
    Display();
}

void Mouse(int button, int state, int x, int y)
{
    if(state == GLUT_DOWN) {
        if(button == GLUT_LEFT_BUTTON) {
            g_mouse_left_down = true;
        }
        else if(button == GLUT_RIGHT_BUTTON) {
            g_mouse_right_down = true;
        }
    }
    else if(state == GLUT_UP) {
        if(button == GLUT_LEFT_BUTTON) {
            g_mouse_left_down = false;
        }
        else if(button == GLUT_RIGHT_BUTTON) {
            g_mouse_right_down = false;
        }
    }
}

void MouseMotion(int x, int y)
{
    // This variable is hack to stop glutWarpPointer from triggering an event callback to Mouse(...)
    // This avoids it being called recursively and hanging up the event loop
    static bool just_warped = false;

    if(just_warped) {
        just_warped = false;
        return;
    }

    if(g_fps_mode) {
        int dx = x - g_viewport_width/2;
        int dy = y - g_viewport_height/2;

        if(dx) {
            g_camera.RotateYaw(g_rotation_speed*dx);
        }

        if(dy) {
            if(g_invert_mouse) {
                g_camera.RotatePitch(g_rotation_speed*dy);
            }
            else {
                g_camera.RotatePitch(-g_rotation_speed*dy);
            }
        }

        glutWarpPointer(g_viewport_width/2, g_viewport_height/2);

        just_warped = true;
    }
}

void DrawHelp()
{
    float y = g_viewport_height - 15;
    float x = 10;
    float ystep = 15;

    glColor3f(0.0f, 1.0f, 0.0f);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    gluOrtho2D(0, g_viewport_width, 0, g_viewport_height);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "SPACEBAR - toggle navigation mode"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "ESC - exit"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "W,A,S,D - move"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "Left/Right mouse button - fly"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "Mouse move - look around"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "1,2,3 - show points/wireframe/texture"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "i - toggle invert mouse"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "t - toggle wireframe sphere"); y -= ystep;
        RenderBitmapString(x, y, 0, GLUT_BITMAP_8_BY_13, "h -  hide this help message"); y -= ystep;

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
}

void RenderBitmapString(float x, float y, float z, void *font, char *string)
{
    char *c;
    glRasterPos3f(x, y,z);

    for (c=string; *c != '\0'; c++) {
        glutBitmapCharacter(font, *c);
    }
}
